/*
   Aclan, Nelson D.
   simulationelson@gmail.com
   Last Modified: September 18, 2009

   An (open) hash table whose cells contain pointers to linked lists;
   collisions and a handful of inputs can be efficiently handled.
*/

#include <stdio.h>
#include <stdlib.h>
#define MAX_VALUE 13

typedef struct StackElem {
   int data;
   struct StackElem *next;
} StackElem;

typedef struct Stack {
   int numElements;
   StackElem *top;
} Stack;

typedef struct OHash {
   int numElements;
   Stack* cell[MAX_VALUE];
} OHash;

void initOHash(OHash *oh);
int isEmptyOH(OHash *oh);
void insert_O(OHash *oh, int key);
void search_O(OHash *oh, int key);
void printOHash(OHash *oh);
void makeEmptyOH(OHash *oh);
void underflow();

void initStack(Stack *s);
void push(Stack *s, int key);

main() {
   OHash oh;
   initOHash(&oh);

   char choice[2];
   int input;

   do {
      printf("-----------------------------------\n");
      printf("|    Open Hash Menu               |\n");
      printf("-----------------------------------\n");
      printf("|    (1) Insert Key               |\n");
      printf("|    (2) Search Key               |\n");
      printf("|    (3) Display All Keys         |\n");
      printf("|    (4) Delete All Keys          |\n");
      printf("|    (5) Exit                     |\n");
      printf("-----------------------------------\n");

      printf("Enter choice: ");
      scanf("%s", choice);

      switch(choice[0]) {
         case '1': /* Insert Key */
            printf("Enter key to be inserted: ");
            scanf("%d", &input);

            if(input <= 0) printf("Invalid key\n");
            else insert_O(&oh, input);
            
            break;
         case '2': /* Search Key */
            if(isEmptyOH(&oh)) {
               underflow();
            } else {
               printf("Enter key to be searched: ");
               scanf("%d", &input);

               if(input <= 0) printf("Invalid key\n");
               else search_O(&oh, input);
            }
            break;
         case '3': /* Display All Keys */
            printOHash(&oh);
            break;
         case '4': /* Delete All Keys */
            makeEmptyOH(&oh);
            break;
         case '5': /* Exit */
            return;
            break;
         default: /* None of the above */
            printf("Bad move\n");
            break;
         }
   } while(choice[0] != '5');
}

/* Initialize open hash table */
void initOHash(OHash *oh) {
   int i;

   oh->numElements = 0;

   /* point each cell to its respective (empty) Stack */
   for(i = 0; i < MAX_VALUE; i++) {
      Stack *s = malloc(sizeof(Stack));
      initStack(s);
      oh->cell[i] = s;
   }
}

/* Check if hash table is empty */
int isEmptyOH(OHash *oh) {
   if(oh->numElements == 0) {
      return 1;
   } else {
      return 0;
   }
}

/* Insert a key into the hash table */
void insert_O(OHash *oh, int key) {
   int hash_index = key % MAX_VALUE;

   push(oh->cell[hash_index], key); /* oh->cell[hash_index] is a Stack */

   printf("Key (%d) inserted\n", key);
   oh->numElements++;
}

/* Search a key in the hash table */
void search_O(OHash *oh, int key) {
   int i, found, hash_index;

   hash_index = key % 13;
   found = 0;

   StackElem *current = oh->cell[hash_index]->top; /* top of the Stack */

   while(current != NULL) {
      if(current->data == key) {
         printf("Key (%d) is found in hash table cell no. %d\n", key, hash_index);
         found = 1; /* not really concerned if the key is unique in that cell */
         break;
      }

      current = current->next;
   }

   if(!found) {
      printf("Key not found\n");
   }
}

/* Print hash table contents */
void printOHash(OHash *oh) {
   if(isEmptyOH(oh)) {
      underflow();
   } else {
      int i;

      /* traverse each cell */
      /* traverse linked list in cell */
      for(i = 0; i < MAX_VALUE; i++) {
         StackElem *current = oh->cell[i]->top;

         if(current != NULL) {
            printf("Cell[%d] = {", i);
            while(current != NULL) {
               printf("%d", current->data);
               current = current->next;

               if(current != NULL) {
                  printf(",");
               }
            }
            printf("}\n");
         }
      }
   }
}

/* Empty the hash table */
void makeEmptyOH(OHash *oh) {
   if(isEmptyOH(oh)) {
      underflow();
   } else {
      int i;

      for(i = 0; i < MAX_VALUE; i++) {
         StackElem *current = oh->cell[i]->top;

         while(current != NULL) {
            StackElem *temp = malloc(sizeof(StackElem));

            temp = oh->cell[i]->top;
            oh->cell[i]->top = oh->cell[i]->top->next; /* Stack->top = Stack->top->next */

            oh->cell[i]->numElements--; /* decrement Stack elements count */
            oh->numElements--; /* decrement hash table elements count */

            current = current->next;
            free(temp);
         }
      }

      printf("Hash Table Emptied\n");
   }
}

/* Nothing much */
void underflow() {
   printf("Hash Table Empty\n");
}

/* Push an element on top of the stack */
void push(Stack *s, int key) {
   StackElem *temp = malloc(sizeof(StackElem));

   temp->data = key;
   temp->next = s->top;
   s->top = temp;
   s->numElements++;
}

/* Initialize empty stack */
void initStack(Stack *s) {
   s->numElements = 0;
   s->top = NULL;
}
